/*
 * Created on Apr 22, 2005
 *
 * Small client API for managing reservations. 
 * Reservation can be created, committed and destroyed. 
 * Also contains convenience methods for loading and storing 
 * EPRs to and from files. 
 *
 */
package se.umu.wsag.client;

import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.Map;
import java.util.HashMap;

import javax.xml.rpc.ServiceException;
import javax.xml.rpc.Stub;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.NCName;
import org.apache.log4j.Logger;
import org.ggf.ws_agreement.Agreement;
import org.ggf.ws_agreement.AgreementContextType;
import org.ggf.ws_agreement.AgreementFactory;
import org.ggf.ws_agreement.AgreementFactoryServiceAddressingLocator;
import org.ggf.ws_agreement.AgreementServiceAddressingLocator;
import org.ggf.ws_agreement.AgreementTemplateType;
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.BusinessValueListType;
import org.ggf.ws_agreement.ContinuingFaultType;
import org.ggf.ws_agreement.GuaranteeTermType;
import org.ggf.ws_agreement.ServiceDescriptionTermType;
import org.ggf.ws_agreement.ServiceNameSet;
import org.ggf.ws_agreement.ServicePropertiesType;
import org.ggf.ws_agreement.TermCompositorType;
import org.ggf.ws_agreement.TermTreeType;
import org.ggf.ws_agreement.VariableSetType;
import org.ggf.ws_agreement.VariableType;
import org.globus.axis.util.Util;
import org.globus.delegation.DelegationConstants;
import org.globus.delegation.DelegationUtil;
import org.globus.gsi.GlobusCredential;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.ResourceNotDestroyedFaultType;
import org.oasis.wsrf.lifetime.SetTerminationTime;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.oasis.wsrf.properties.ResourceUnknownFaultType;

import se.umu.reservation.EarliestStartType;
import se.umu.reservation.LatestStartType;
import se.umu.reservation.ReservationDescriptionType;
import se.umu.reservation.ReservationServiceLevelObjective;
import se.umu.reservation.ReservationType;
import se.umu.wsag.impl.AgreementConstants;
import se.umu.wsag.impl.AgreementFactoryConstants;
import se.umu.wsag.util.JobSubmissionUtils;


/**
 * @author tordsson
 *
 * ReservationClient is a client library for handling reservation (agreements).  
 * Reservations can be created, cancelled, committed (no longer soft-state), 
 * and queried for resource properties.    
 */
public class ReservationClient {
	
	private static Logger logger = Logger.getLogger(ReservationClient.class); 
	
	private Map factories; 
	private AgreementInitiator initiator; 
	
    static {
        Util.registerTransport(); 
    }
    

    public ReservationClient() { 
    	this.factories = new HashMap(); 
    	this.initiator = new AgreementInitiator(); 
    }

    /**
     * Create a new 
     * @param factoryURL
     * @return
     * @throws Exception
     */
	public AgreementFactory newAgreementFactory(String factoryURL) 
	throws Exception {
        AgreementFactoryServiceAddressingLocator locator = 
        new AgreementFactoryServiceAddressingLocator();             

        EndpointReferenceType endpoint = new EndpointReferenceType();
        endpoint.setAddress(new Address(factoryURL));
        AgreementFactory factory = locator.getAgreementFactoryPort(endpoint);
        
    	// Client security descriptor file     		
        String clientDesc = "etc/wsag/client-security-config.xml"; 
        // Set descriptor on Stub             
        ((Stub)factory)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				clientDesc);  
     
        return factory; 
	}
    
    
    /**
     * Delegate user credential to container running AgreementFactoryService
     * @param agreementHost
     * @param securityDescriptor
     * @return
     * @throws Exception
     */
    public EndpointReferenceType delegateCredentials(String agreementHost, 
    		ClientSecurityDescriptor securityDescriptor) throws Exception {
    	
    	ServiceURL agreementFactoryURL = new ServiceURL(agreementHost); 
    	String hostName = agreementFactoryURL.getURL().getHost(); 
    	String URL = ServiceURL.getURL(hostName, 
    			DelegationConstants.FACTORY_PATH).toString(); 
    	
    	logger.debug("Delegating credential to:  " + URL); 
    	EndpointReferenceType epr = new EndpointReferenceType(); 
    	epr.setAddress(new Address(URL)); 
    	
    	X509Certificate[] certs =  DelegationUtil.getCertificateChainRP(epr,
    			securityDescriptor); // pass desc instead of null
    	
    	if(certs == null) {
    		throw new Exception("getCertificateChainRP failed"); 
    	}
    	else {
    
    		// Get the credential delegated from client to JSS, 
    		// and delegate it further using DelegationService... 
    		GlobusCredential credential = JobSubmissionUtils.
					getDelegatedCredential(); 
    		// GlobusCredential credential = GlobusCredential.getDefaultCredential(); 
    		EndpointReferenceType delegationEpr = DelegationUtil.delegate(
    				URL, credential, certs[0], true, securityDescriptor); 
    		
    		if(delegationEpr == null)
    			logger.debug("Delegation failed"); 
    		
    		return delegationEpr; 
    	}
    }
    
    /**
     * @param epr Reference to the WS-resource representing the reservation
     * @return The resource properties of the reservation. 
     * @throws Exception
     */
    public ReservationType getReservationProperties(EndpointReferenceType epr)
    	throws Exception {
    		
    	AgreementServiceAddressingLocator locator = 
    		new AgreementServiceAddressingLocator();
    	
    	Agreement agreement = locator.getAgreementPort(epr);
    	
    	// Get AgreementTerms   
    	GetResourcePropertyResponse termsResponse = 
    		agreement.getResourceProperty(AgreementConstants.TERMS_RP);
    	MessageElement any[] = termsResponse.get_any();
    	               //any = termsResponse.get_any();
    	
    	// 1: Extract TermTree        
    	TermTreeType terms = (TermTreeType) ObjectDeserializer.toObject(
    			any[0], TermTreeType.class); 
    	
    	any = terms.getAll().getServiceDescriptionTerm().get_any();         
    	ReservationType reservationInfo = (ReservationType) 
			ObjectDeserializer.toObject(any[0], ReservationType.class); 
    	
    	return reservationInfo; 
    	
    }
 
    
    /**
     * Helper method for createReservation, retrives all 
     * agreement templates and returns the one for reservations. 
     * Contacts 
     * @return
     * @throws Exception
     */
    private AgreementTemplateType getTemplate(AgreementInitiator initiator, 
    	AgreementFactory factory) 
			throws Exception {    				
		
		AgreementTemplateType templates[] = initiator.getTemplates(factory); 
		
		AgreementTemplateType tmpTemplate;
		AgreementTemplateType reservationTemplate = null;
		for(int i = 0; i < templates.length; i++) {
			
			tmpTemplate = templates[i]; 
			String name= tmpTemplate.getName().toString(); 

			if("ReservationTemplate".equalsIgnoreCase(name)) {			
				reservationTemplate = tmpTemplate; 							
			}			
		}
		
		if(reservationTemplate == null)
			throw new Exception("Template \"ReservationTemplate\" not found!");
		
		return reservationTemplate; 
    
    }
    
    
    /**
     * Helper method to createReservation
     * Fills out the template with the terms we specified to create 
     * an agreement offer (that will be sent to the factory) 
     * @param template
     * @param cpus
     * @param dur
     * @param earliest
     * @param latest
     * @param flex
     * @return
     * @throws Exception
     */
    private AgreementType createOffer(AgreementTemplateType template, 
    		int numberOfCPUs, 
			int duration, // (in seconds) 
			Calendar earliest, 
			Calendar latest, 
			boolean flexible, 
			EndpointReferenceType delegatedCredentialEpr) throws Exception {
    	
    	AgreementType offer = new AgreementType();
    	
    	NCName name = new NCName("ReservationOffer");      	
    	offer.setName(name);    	
    	
    	// Set epr containing delegated creds as agreementinitiator. 
    	AgreementContextType context = new AgreementContextType(); 
    	context.setAgreementInitiator(delegatedCredentialEpr); 
    	offer.setContext(context);     	
    	
    	NCName reservationName = new NCName("Reservation"); 	    	
    	TermTreeType termsTree = new TermTreeType();
    	TermCompositorType all = new TermCompositorType();    
    	
    	// 1:  
    	// ServiceDescriptionTerms, these are added to an Object
    	// (ReservationDescriptionType) due to axis bug 15789
    	// resulting in only one ServiceDescriptionTerm in the Agreement
    	ReservationDescriptionType description = 
    		new ReservationDescriptionType();

		description.setDuration(duration); 
    	description.setNumberOfCPUs(numberOfCPUs);
    	description.setFlexible(flexible); 
	
    	ServiceDescriptionTermType serviceDescriptionTerm = 
    		new ServiceDescriptionTermType();   
    	MessageElement descriptionElement[] = AnyHelper.toAnyArray(description); 
    	serviceDescriptionTerm.set_any(descriptionElement);
    	
       	all.setServiceDescriptionTerm(serviceDescriptionTerm);     	
       	
       	// 2: Set up service properties (just defines names for 
       	// usage in GuaranteeTerms), 
       	// "earlyStart" and "lateStart" are defined
       	ServicePropertiesType servicePropertiesType = 
       		new ServicePropertiesType();
       	servicePropertiesType.setName(reservationName); 
       	
       	VariableSetType variableSetType = new VariableSetType();
       	
       	VariableType earliestStartVariable = new VariableType();
       	earliestStartVariable.setName(new NCName("earlyStart")); 
       	earliestStartVariable.setMetric(
       			EarliestStartType.getTypeDesc().getXmlType());
       	
       	VariableType latestStartVariable = new VariableType(); 
       	latestStartVariable.setName(new NCName("lateStart"));  
       	latestStartVariable.setMetric(
       			LatestStartType.getTypeDesc().getXmlType());
       	
       	VariableType variables[] = new VariableType[] {
       		earliestStartVariable, latestStartVariable};        
       	variableSetType.setVariable(variables);       	       
       	servicePropertiesType.setVariableSet(variableSetType);
       	
       	all.setServiceProperties(servicePropertiesType); 
       	
       	// 3: GuaranteeTerms 
       	// (things are getting a bit more complicated here)
       	
    	// First, add serviceScope, BusinessValueList and Name, none of 
       	// these are used by the AgreementFactoryService in a reservation 
       	// scenario. 
    	GuaranteeTermType guaranteeTerm = new GuaranteeTermType();
    	
		ServiceNameSet serviceScope = new ServiceNameSet();
		NCName serviceNames[] = new NCName[]{reservationName}; 
    	serviceScope.setServiceName(serviceNames);
    	guaranteeTerm.setServiceScope(serviceScope);         	
    	BusinessValueListType businessValueList = new BusinessValueListType();
    	guaranteeTerm.setBusinessValueList(businessValueList);    	
    	guaranteeTerm.setName(reservationName); 
       	 
    	
    	// Next, create ServiceLevelObjectiveType and fill it with a
    	// ReservationServiceLevelObjectiveType
    	ReservationServiceLevelObjective SLOTerms = 
    		new ReservationServiceLevelObjective();    	
    	String terms[] = new String[2];     	
    	
     	terms[0] = "earlyStart IS_AFTER " + (earliest.getTimeInMillis() / 1000);     
    	terms[1] = "lateStart IS_BEFORE " + (latest.getTimeInMillis() / 1000);
    	
    	SLOTerms.setTerm(terms);     
    	
    	// After registring a typemapping in server-config.wsdd 
    	// it works fine to pass a ReservationServiceLevelObjective
    	// in the guaranteeterm.  
        guaranteeTerm.setServiceLevelObjective(SLOTerms);
    	all.setGuaranteeTerm(guaranteeTerm); 
     	
    	// Add ServiceDescriptionTerms and GuaranteeTerms to agreement.  
    	termsTree.setAll(all); 
    	offer.setTerms(termsTree);  
    	
    	return offer;  
    	    
    }
    
    /**
     * Creates a reservation. Contacts AgreementFactoryService
     * and retrieves list of AgreementTemplates. Selects the one 
     * for requesting reservations. Fills out the template with the
     * five parameters (numberOfCPUs, earliestStart, latestStart, 
     * duration, flexible), and sends the requests to the factory. 
     * Upon a successful request, an agreement epr is returned.  
     * @param factoryURI
     * @param numberOfCPUs
     * @param earliestStart
     * @param latestStart
     * @param duration in minutes. 
     * @param flexible
     * @return
     */             
    public EndpointReferenceType createReservation(String factoryURL, 
    		int numberOfCPUs, 
			Calendar earliestStart, 
			Calendar latestStart, 
			int duration, 
			boolean flexible) throws Exception {
    	
    	try {
    		
    		// Initiator can perform two basic tasks, 
    		// fetching agreement templates, and
    		// sending agreement request to factory.     		
    		logger.debug("Create reservation, getting initiator..."); 
    		// AgreementInitiator initiator = new AgreementInitiator(factoryURI);
    		// AgreementFactory factory = initiator.getFactory(factoryURI);
    		
    		logger.debug("Create reservation, setting security options"); 
    		
    		// Get AgreementFactory
    		boolean newFactory = false; 
    		AgreementFactory factory = (AgreementFactory) factories.get(factoryURL); 
    		if(factory == null) {
    			factory = newAgreementFactory(factoryURL); 
    			factories.put(factoryURL, factory); 
    			newFactory = true; 
    		}
    	
            logger.debug("Getting templates"); 
    		// Get all templates, and selected the one specifying reservations. 
        	AgreementTemplateType reservationTemplate = getTemplate(
        			initiator, factory);        	            
        	
        	// TODO: Should be able to save the delegatedCredentialEPR somewhere...
        	// if(newFactory) {
        		// Delegate Credentials:      
        		String clientDesc = "etc/wsag/client-security-config.xml"; 
        		ClientSecurityDescriptor desc = ClientSecurityConfig.initialize(
        				clientDesc);     
        		EndpointReferenceType delegatedCredentialEpr = delegateCredentials(
        				factoryURL, desc);        
        		if(delegatedCredentialEpr == null) {
        			throw new Exception("Credential delegation failed"); 
        		}
        	// }
        	
        	// Fill in template with our requested parameters.   
    		AgreementType offer = createOffer(reservationTemplate, 
        			numberOfCPUs, 
    				duration, 
    				earliestStart, 
    				latestStart, 
    				flexible, 
					delegatedCredentialEpr); 
    		
    		// Send offer to factory 
    	  	EndpointReferenceType epr = initiator.createAgreement(factory, offer);
    		
    		return epr; 
    		
    	} catch (ContinuingFaultType f) {
    		logger.error("Creation failed: ");
    		FaultHelper helper = new FaultHelper(f); 
    		logger.error("error message: " + helper.getDescriptionAsString()); 
    		throw f; 
    		// throw new Exception(helper.getDescriptionAsString()); 
    	}
		catch (Exception e) {
    		logger.error("Creation failed: " + e.getMessage()); 
    		throw e;  
    	}
    }
    
    
    /**
     * Destroys a created reservation by removing the resource 
     * representing it. 
     * @param epr
     * @throws ResourceNotDestroyedFaultType
     * @throws ResourceUnknownFaultType 
     * @throws ServiceException 
     */        
    public void destroyReservation(EndpointReferenceType epr)  
    	throws Exception {
    	
        AgreementServiceAddressingLocator locator = 
        	new AgreementServiceAddressingLocator();        
        Agreement agreement = locator.getAgreementPort(epr); 
        agreement.destroy(new Destroy());     	
    }

    /**
     * Commits a reservation, that is, extends lifetime to ensure
     * that the agreement will exist until the reservation has ended. 
     * Unless this is done, the agreement times out shortly after its
     * creations. 
     * @param epr
     */
    public void commitReservation(EndpointReferenceType epr) throws Exception {
    	
    	logger.debug("Committing reservation, that is extending resource lifetime"); 
    	
    	AgreementServiceAddressingLocator locator = 
    		new AgreementServiceAddressingLocator();
    	
    	Agreement agreement = locator.getAgreementPort(epr);
    	
    	ReservationType reservationInfo = getReservationProperties(epr); 
    	// Let termination time be startTime + duration + 1 minute (margin)
    	Calendar startTime = reservationInfo.getExactStart();  
		int duration = reservationInfo.getDuration(); 
    	
    	// Add an extra minute just in case...
    	startTime.add(Calendar.SECOND, duration + 60);
    	logger.debug("Terminating agreement resource at: " 
    			+ startTime.getTime());
    	
    	SetTerminationTime request = new SetTerminationTime(startTime);         
    	agreement.setTerminationTime(request);    	
    	
    }
    
    /**
     * Modify reservation. As this operation is not supported by the 
     * WSAG specification, we have to settle with a non-atomic implementation. 
     * First remove old reservation, next create a new one with the new terms. 
     * 
     * @param epr Reference to existing reservation. 
     * @param numberOfCPUs
     * @param earliestStart
     * @param latestStart
     * @param duration
     * @param flexible
     * @throws Exception
     */
    public EndpointReferenceType modifyReservation(EndpointReferenceType epr,
    		int numberOfCPUs, 
			Calendar earliestStart, 
			Calendar latestStart, 
			int duration, 
			boolean flexible) throws Exception {
    	
    	try { 
    		
    		// Save URL to reservationService: 
    		String factoryHost = epr.getAddress().getHost(); 
    	   	String factoryURL = ServiceURL.getURL(factoryHost, 
        			AgreementFactoryConstants.serviceName).toString(); 
    		logger.warn("Trying to create (modify) agreement using: " + factoryURL); 
    		
    		// Remove old reservation
    		destroyReservation(epr); 

    		// Create new reservation
    	    EndpointReferenceType newEpr = createReservation(factoryURL, 
    	    		numberOfCPUs,   
    				earliestStart, 
    				latestStart, 
    				duration, 
    				flexible); 
    	    
    	    return newEpr; 
    	} catch (ContinuingFaultType f) {
    		// Modification not possible (but maybe possible later) 
    		throw f; 
    	} catch (Exception e) {
    		// General failure 
    		throw e; 
		
    	}
    }

}
